@@ -116,12 +116,29 @@ sub currstr {
 }
 
 package version::vpp;
-use strict;
 
-use POSIX qw/locale_h/;
-use locale;
-use vars qw ($VERSION @ISA @REGEXS);
-$VERSION = 0.99;
+use 5.006002;
+use strict;
+use warnings::register;
+
+use Config;
+
+our $VERSION = 0.9924;
+our $CLASS = 'version::vpp';
+our ($LAX, $STRICT, $WARN_CATEGORY);
+
+if ($] > 5.015) {
+    warnings::register_categories(qw/version/);
+    $WARN_CATEGORY = 'version';
+} else {
+    $WARN_CATEGORY = 'numeric';
+}
+
+require version::regex;
+*version::vpp::is_strict = \&version::regex::is_strict;
+*version::vpp::is_lax = \&version::regex::is_lax;
+*LAX = \$version::regex::LAX;
+*STRICT = \$version::regex::STRICT;
 
 use overload (
     '""'       => \&stringify,
@@ -140,14 +157,62 @@ use overload (
     'abs'      => \&vnoop,
 );
 
-eval "use warnings";
-if ($@) {
-    eval '
-	package
-	warnings;
-	sub enabled {return $^W;}
-	1;
-    ';
+sub import {
+    no strict 'refs';
+    my ($class) = shift;
+
+    # Set up any derived class
+    unless ($class eq $CLASS) {
+	local $^W;
+	*{$class.'::declare'} =  \&{$CLASS.'::declare'};
+	*{$class.'::qv'} = \&{$CLASS.'::qv'};
+    }
+
+    my %args;
+    if (@_) { # any remaining terms are arguments
+	map { $args{$_} = 1 } @_
+    }
+    else { # no parameters at all on use line
+	%args =
+	(
+	    qv => 1,
+	    'UNIVERSAL::VERSION' => 1,
+	);
+    }
+
+    my $callpkg = caller();
+
+    if (exists($args{declare})) {
+	*{$callpkg.'::declare'} =
+	    sub {return $class->declare(shift) }
+	  unless defined(&{$callpkg.'::declare'});
+    }
+
+    if (exists($args{qv})) {
+	*{$callpkg.'::qv'} =
+	    sub {return $class->qv(shift) }
+	  unless defined(&{$callpkg.'::qv'});
+    }
+
+    if (exists($args{'UNIVERSAL::VERSION'})) {
+	no warnings qw/redefine/;
+	*UNIVERSAL::VERSION
+		= \&{$CLASS.'::_VERSION'};
+    }
+
+    if (exists($args{'VERSION'})) {
+	*{$callpkg.'::VERSION'} = \&{$CLASS.'::_VERSION'};
+    }
+
+    if (exists($args{'is_strict'})) {
+	*{$callpkg.'::is_strict'} = \&{$CLASS.'::is_strict'}
+	  unless defined(&{$callpkg.'::is_strict'});
+    }
+
+    if (exists($args{'is_lax'})) {
+	*{$callpkg.'::is_lax'} = \&{$CLASS.'::is_lax'}
+	  unless defined(&{$callpkg.'::is_lax'});
+    }
 }
 
 my $VERSION_MAX = 0x7FFFFFFF;
@@ -271,6 +336,7 @@ dotted_decimal_version:
     } 					# end if dotted-decimal
     else
     {					# decimal versions
+	my $j = 0;
 	# special $strict case for leading '.' or '0'
 	if ($strict) {
 	    if ($d eq '.') {
@@ -333,7 +399,7 @@ dotted_decimal_version:
 	}
 
 	while (isDIGIT($d)) {
-	    $d++;
+	    $d++; $j++;
 	    if ($d eq '.' && isDIGIT($d-1)) {
 		if ($alpha) {
 		    return BADVERSION($s,$errstr,"Invalid version format (underscores before decimal)");
@@ -355,6 +421,7 @@ dotted_decimal_version:
 		if ( ! isDIGIT($d+1) ) {
 		    return BADVERSION($s,$errstr,"Invalid version format (misplaced underscore)");
 		}
+		$width = $j;
 		$d++;
 		$alpha = TRUE;
 	    }
@@ -370,6 +437,10 @@ version_prescan_finish:
 	# trailing non-numeric data
 	return BADVERSION($s,$errstr,"Invalid version format (non-numeric data)");
     }
+    if ($saw_decimal > 1 && ($d-1) eq '.') {
+	# no trailing period allowed
+	return BADVERSION($s,$errstr,"Invalid version format (trailing decimal)");
+    }
 
     if (defined $sqv) {
 	$$sqv = $qv;
@@ -410,7 +481,7 @@ sub scan_version {
     if ($errstr) {
 	# 'undef' is a special case and not an error
 	if ( $s ne 'undef') {
-	    use Carp;
+	    require Carp;
 	    Carp::croak($errstr);
 	}
     }
@@ -431,7 +502,7 @@ sub scan_version {
 	$$rv->{width} = $width;
     }
 
-    while (isDIGIT($pos)) {
+    while (isDIGIT($pos) || $pos eq '_') {
 	$pos++;
     }
     if (!isALPHA($pos)) {
@@ -452,6 +523,7 @@ sub scan_version {
  		if ( !$qv && $s > $start && $saw_decimal == 1 ) {
 		    $mult *= 100;
  		    while ( $s < $end ) {
+			next if $s eq '_';
 			$orev = $rev;
  			$rev += $s * $mult;
  			$mult /= 10;
@@ -471,6 +543,7 @@ sub scan_version {
   		}
  		else {
  		    while (--$end >= $s) {
+			next if $end eq '_';
 			$orev = $rev;
  			$rev += $end * $mult;
  			$mult *= 10;
@@ -508,7 +581,7 @@ sub scan_version {
 		last;
 	    }
 	    if ( $qv ) {
-		while ( isDIGIT($pos) ) {
+		while ( isDIGIT($pos) || $pos eq '_') {
 		    $pos++;
 		}
 	    }
@@ -566,21 +639,49 @@ sub scan_version {
     return $s;
 }
 
-sub new
-{
-	my ($class, $value) = @_;
-	my $self = bless ({}, ref ($class) || $class);
-	my $qv = FALSE;
+sub new {
+    my $class = shift;
+    unless (defined $class or $#_ > 1) {
+	require Carp;
+	Carp::croak('Usage: version::new(class, version)');
+    }
 
-	if ( ref($value) && eval('$value->isa("version")') ) {
-	    # Can copy the elements directly
-	    $self->{version} = [ @{$value->{version} } ];
-	    $self->{qv} = 1 if $value->{qv};
-	    $self->{alpha} = 1 if $value->{alpha};
-	    $self->{original} = ''.$value->{original};
-	    return $self;
-	}
+    my $self = bless ({}, ref ($class) || $class);
+    my $qv = FALSE;
 
+    if ( $#_ == 1 ) { # must be CVS-style
+	$qv = TRUE;
+    }
+    my $value = pop; # always going to be the last element
+
+    if ( ref($value) && eval('$value->isa("version")') ) {
+	# Can copy the elements directly
+	$self->{version} = [ @{$value->{version} } ];
+	$self->{qv} = 1 if $value->{qv};
+	$self->{alpha} = 1 if $value->{alpha};
+	$self->{original} = ''.$value->{original};
+	return $self;
+    }
+
+    if ( not defined $value or $value =~ /^undef$/ ) {
+	# RT #19517 - special case for undef comparison
+	# or someone forgot to pass a value
+	push @{$self->{version}}, 0;
+	$self->{original} = "0";
+	return ($self);
+    }
+
+
+    if (ref($value) =~ m/ARRAY|HASH/) {
+	require Carp;
+	Carp::croak("Invalid version format (non-numeric data)");
+    }
+
+    $value = _un_vstring($value);
+
+    if ($Config{d_setlocale}) {
+	use POSIX qw/locale_h/;
+	use if $Config{d_setlocale}, 'locale';
 	my $currlocale = setlocale(LC_ALL);
 
 	# if the current locale uses commas for decimal points, we
@@ -589,108 +690,69 @@ sub new
 	if ( localeconv()->{decimal_point} eq ',' ) {
 	    $value =~ tr/,/./;
 	}
+    }
 
-	if ( not defined $value or $value =~ /^undef$/ ) {
-	    # RT #19517 - special case for undef comparison
-	    # or someone forgot to pass a value
-	    push @{$self->{version}}, 0;
-	    $self->{original} = "0";
-	    return ($self);
-	}
+    # exponential notation
+    if ( $value =~ /\d+.?\d*e[-+]?\d+/ ) {
+	$value = sprintf("%.9f",$value);
+	$value =~ s/(0+)$//; # trim trailing zeros
+    }
 
-	if ( $#_ == 2 ) { # must be CVS-style
-	    $value = $_[2];
-	    $qv = TRUE;
-	}
+    my $s = scan_version($value, \$self, $qv);
 
-	$value = _un_vstring($value);
+    if ($s) { # must be something left over
+	warn(sprintf "Version string '%s' contains invalid data; "
+		   ."ignoring: '%s'", $value, $s);
+    }
 
-	# exponential notation
-	if ( $value =~ /\d+.?\d*e[-+]?\d+/ ) {
-	    $value = sprintf("%.9f",$value);
-	    $value =~ s/(0+)$//; # trim trailing zeros
-	}
-
-	my $s = scan_version($value, \$self, $qv);
-
-	if ($s) { # must be something left over
-	    warn("Version string '%s' contains invalid data; "
-                       ."ignoring: '%s'", $value, $s);
-	}
-
-	return ($self);
+    return ($self);
 }
 
 *parse = \&new;
 
-sub numify
-{
+sub numify {
     my ($self) = @_;
     unless (_verify($self)) {
 	require Carp;
 	Carp::croak("Invalid version object");
     }
-    my $width = $self->{width} || 3;
     my $alpha = $self->{alpha} || "";
     my $len = $#{$self->{version}};
     my $digit = $self->{version}[0];
     my $string = sprintf("%d.", $digit );
 
-    for ( my $i = 1 ; $i < $len ; $i++ ) {
-	$digit = $self->{version}[$i];
-	if ( $width < 3 ) {
-	    my $denom = 10**(3-$width);
-	    my $quot = int($digit/$denom);
-	    my $rem = $digit - ($quot * $denom);
-	    $string .= sprintf("%0".$width."d_%d", $quot, $rem);
-	}
-	else {
-	    $string .= sprintf("%03d", $digit);
-	}
+    if ($alpha and warnings::enabled()) {
+	warnings::warn($WARN_CATEGORY, 'alpha->numify() is lossy');
     }
 
-    if ( $len > 0 ) {
-	$digit = $self->{version}[$len];
-	if ( $alpha && $width == 3 ) {
-	    $string .= "_";
-	}
-	$string .= sprintf("%0".$width."d", $digit);
+    for ( my $i = 1 ; $i <= $len ; $i++ ) {
+	$digit = $self->{version}[$i];
+	$string .= sprintf("%03d", $digit);
     }
-    else # $len = 0
-    {
+
+    if ( $len == 0 ) {
 	$string .= sprintf("000");
     }
 
     return $string;
 }
 
-sub normal
-{
+sub normal {
     my ($self) = @_;
     unless (_verify($self)) {
 	require Carp;
 	Carp::croak("Invalid version object");
     }
-    my $alpha = $self->{alpha} || "";
+
     my $len = $#{$self->{version}};
     my $digit = $self->{version}[0];
     my $string = sprintf("v%d", $digit );
 
-    for ( my $i = 1 ; $i < $len ; $i++ ) {
+    for ( my $i = 1 ; $i <= $len ; $i++ ) {
 	$digit = $self->{version}[$i];
 	$string .= sprintf(".%d", $digit);
     }
 
-    if ( $len > 0 ) {
-	$digit = $self->{version}[$len];
-	if ( $alpha ) {
-	    $string .= sprintf("_%0d", $digit);
-	}
-	else {
-	    $string .= sprintf(".%0d", $digit);
-	}
-    }
-
     if ( $len <= 2 ) {
 	for ( $len = 2 - $len; $len != 0; $len-- ) {
 	    $string .= sprintf(".%0d", 0);
@@ -700,8 +762,7 @@ sub normal
     return $string;
 }
 
-sub stringify
-{
+sub stringify {
     my ($self) = @_;
     unless (_verify($self)) {
 	require Carp;
@@ -714,9 +775,7 @@ sub stringify
 	    : $self->numify;
 }
 
-sub vcmp
-{
-    require UNIVERSAL;
+sub vcmp {
     my ($left,$right,$swap) = @_;
     my $class = ref($left);
     unless ( UNIVERSAL::isa($right, $class) ) {
@@ -746,20 +805,6 @@ sub vcmp
 	$i++;
     }
 
-    # tiebreaker for alpha with identical terms
-    if ( $retval == 0
-	&& $l == $r
-	&& $left->{version}[$m] == $right->{version}[$m]
-	&& ( $lalpha || $ralpha ) ) {
-
-	if ( $lalpha && !$ralpha ) {
-	    $retval = -1;
-	}
-	elsif ( $ralpha && !$lalpha) {
-	    $retval = +1;
-	}
-    }
-
     # possible match except for trailing 0's
     if ( $retval == 0 && $l != $r ) {
 	if ( $l < $r ) {
@@ -800,7 +845,7 @@ sub is_alpha {
 
 sub qv {
     my $value = shift;
-    my $class = 'version';
+    my $class = $CLASS;
     if (@_) {
 	$class = ref($value) || $value;
 	$value = shift;
@@ -808,7 +853,7 @@ sub qv {
 
     $value = _un_vstring($value);
     $value = 'v'.$value unless $value =~ /(^v|\d+\.\d+\.\d)/;
-    my $obj = version->new($value);
+    my $obj = $CLASS->new($value);
     return bless $obj, $class;
 }
 
@@ -847,16 +892,16 @@ sub _is_non_alphanumeric {
 sub _un_vstring {
     my $value = shift;
     # may be a v-string
-    if ( length($value) >= 3 && $value !~ /[._]/
+    if ( length($value) >= 1 && $value !~ /[,._]/
 	&& _is_non_alphanumeric($value)) {
 	my $tvalue;
-	if ( $] ge 5.008_001 ) {
+	if ( $] >= 5.008_001 ) {
 	    $tvalue = _find_magic_vstring($value);
 	    $value = $tvalue if length $tvalue;
 	}
-	elsif ( $] ge 5.006_000 ) {
+	elsif ( $] >= 5.006_000 ) {
 	    $tvalue = sprintf("v%vd",$value);
-	    if ( $tvalue =~ /^v\d+(\.\d+){2,}$/ ) {
+	    if ( $tvalue =~ /^v\d+(\.\d+)*$/ ) {
 		# must be a v-string
 		$value = $tvalue;
 	    }
@@ -881,6 +926,7 @@ sub _find_magic_vstring {
 	    $magic = $magic->MOREMAGIC;
 	}
     }
+    $tvalue =~ tr/_//d;
     return $tvalue;
 }
 
